import hashlib

from serpent import Serpent
from serpent import serpent_cbc_encrypt 
from serpent import serpent_cbc_decrypt 
from commands import BROADCAST
from commands import DIRECT 
from commands import GETDATA 
from commands import IGNORE
from commands import COMMAND_LABELS
import base64
import binascii
import time
import struct
import hmac
import random
import os 
import logging

PEST_VERSION = 0xFC
PACKET_SIZE = 496 
MAX_SPEAKER_SIZE = 32
TS_ACCEPTABLE_SKEW = 60 * 15
BLACK_PACKET_FORMAT = "<448s48s"
RED_PACKET_FORMAT = "<16sBBxB428s"
RED_PACKET_LENGTH_WITH_PADDING = 448
MESSAGE_PACKET_FORMAT = "<q32s32s32s324s"
GETDATA_MESSAGE_PACKET_FORMAT = "<q32s32s32s32s292s"
PROD_MESSAGE_PACKET_FORMAT = "hB632s32s32s220s"
MAX_MESSAGE_LENGTH = 428
NONCE_LENGTH = 16
TEXT_PAYLOAD_SIZE = 324

# error codes
STALE_PACKET = 0
DUPLICATE_PACKET = 1
MALFORMED_PACKET = 2
INVALID_SIGNATURE = 3
UNSUPPORTED_VERSION = 4
OUT_OF_ORDER_SELF = 5
OUT_OF_ORDER_NET = 6
OUT_OF_ORDER_BOTH = 7
INVALID_HANDLE_ENCODING = 8

# logging formats
OUTGOING_MESSAGE_LOGGING_FORMAT = u"[%s:%d %s] <- %s %s %s %s"
INCOMING_MESSAGE_LOGGING_FORMAT = u"[%s:%d %s] -> %s %s %d %s"

# fork status values
FORKED = 0
NOT_FORKED = 1
FIRST_MESSAGE = 2
EMPTY_CHAIN = "\x00" * 32 

class Message(object):
    def __init__(self, message, state=None):
        self.state = state
        self.prefix = None
        self.fork_status = None
        # target peer handle
        self.handle = message.get("handle")
        self.peer = message.get("peer")
        self.body = message.get("body")
        self.timestamp = message.get("timestamp")
        self.command = message.get("command")
        # source peer handle
        self.speaker = message.get("speaker")
        self.bounces = message.get("bounces")
        self.self_chain = message.get("self_chain")
        self.net_chain = message.get("net_chain")
        self.self_chain_valid = message.get("self_chain_valid")
        self.error_code = message.get("error_code")
        self.message_hash = message.get("message_hash")
        self.message_bytes = message.get("message_bytes")
        self.long_buffer = message.get("long_buffer")
        self.get_data_response = message.get("get_data_response")
        self.metadata = message.get('metadata')
        self.original = message.get('original')
        self.reporting_peers = message.get('reporting_peers', [])
        self.warning = message.get('warning')
        self.order_buffer = message.get('order_buffer')
        self.error_code = message.get('error_code')

    @classmethod
    def pack(cls, peer, command, bounces, message_bytes):
        key_bytes = base64.b64decode(peer.get_key())
        signing_key = key_bytes[:32]
        cipher_key = key_bytes[32:]

        # pack packet bytes
        nonce = os.urandom(NONCE_LENGTH)

        version = PEST_VERSION
        red_packet_bytes = struct.pack(RED_PACKET_FORMAT,
                                       nonce,
                                       bounces,
                                       version,
                                       command,
                                       cls._pad(message_bytes, MAX_MESSAGE_LENGTH))

        # encrypt packet
        serpent = Serpent(cipher_key)
        black_packet_bytes = serpent_cbc_encrypt(cipher_key, red_packet_bytes)

        # sign packet
        signature_bytes = hmac.new(signing_key, black_packet_bytes, hashlib.sha384).digest()

        # pack the signed black packet
        signed_packet_bytes = struct.pack(BLACK_PACKET_FORMAT, black_packet_bytes, signature_bytes)

        return signed_packet_bytes

    @classmethod
    def unpack(cls, peer, black_packet, long_buffer, order_buffer, metadata, state=None):
        # unpack the black packet
        for key in peer.keys:
            key_bytes = base64.b64decode(key)
            signing_key = key_bytes[:32]
            cipher_key = key_bytes[32:]

            try:
                black_packet_bytes, signature_bytes = struct.unpack(BLACK_PACKET_FORMAT,
                                                                    black_packet)
            except:
                logging.error("Discarding malformed black packet from %s" % peer.get_key())
                return {
                    "error_code": MALFORMED_PACKET,
                    "metadata": metadata
                }

            # check signature
            signature_check_bytes = hmac.new(signing_key,
                                             black_packet_bytes,
                                             hashlib.sha384).digest()

            if(signature_check_bytes != signature_bytes):
                continue

            # try to decrypt black packet
            Serpent(cipher_key)
            red_packet_bytes = serpent_cbc_decrypt(cipher_key, black_packet_bytes)

        # unpack red packet
        try:
            nonce, bounces, version, command, message_bytes = struct.unpack(
                RED_PACKET_FORMAT,
                red_packet_bytes
            )

        # if red_packet_bytes was never set, no matching key for the sig was found
        # this is expected to happen often as only one peer's key should match for each
        # message we receive
        except NameError, ex:
            return {
                "error_code": INVALID_SIGNATURE,
                "metadata": metadata
            }

        if version < PEST_VERSION:
            return {
                "error_code": UNSUPPORTED_VERSION,
                "metadata": metadata
            }

        # compute message_hash
        message_hash = cls.gen_hash(message_bytes)
        message = cls._build_message_dict(command,
                                          peer,
                                          bounces,
                                          message_hash,
                                          message_bytes,
                                          metadata)

        if not cls._check_for_ascii_encoding(message['speaker']):
            return {
                "error_code": INVALID_HANDLE_ENCODING,
                "metadata": metadata
            }
        return cls._evaluate_message(long_buffer, order_buffer, message)

    @classmethod
    def _evaluate_message(cls, long_buffer, order_buffer, message):
        # if we're expecting this message as a GETDATA response, skip the timestamp check
        if not order_buffer.expects(message['message_hash']):
            if not cls.in_time_window(message['timestamp']):
                message['error_code'] = STALE_PACKET
                return message
        else:
            # we need to mark this as a get data response in order not to rebroadcast it
            # TODO how to handle out of order unrequested but not stale messages subsequent to a message in the order buffer?
            message['get_data_response'] = True

        if long_buffer.has(message['message_hash']):
            message['error_code'] = DUPLICATE_PACKET
            return message

        if DIRECT is message['command']:
           return cls._evaluate_direct(long_buffer, message)
        elif BROADCAST is message['command']:
            return cls._evaluate_broadcast(long_buffer, message)
        else:
            return message

    @classmethod
    def _build_message_dict(cls, command, peer, bounces, message_hash, message_bytes, metadata):
        # unpack message
        if command is GETDATA:
            int_ts, self_chain, net_chain, speaker, body = cls._unpack_getdata_message(message_bytes)
        else:
            int_ts, self_chain, net_chain, speaker, body = cls._unpack_message(message_bytes)

        message = {
            "peer": peer,
            "body": body,
            "timestamp": int_ts,
            "command": command,
            "speaker": speaker,
            "bounces": bounces,
            "self_chain": self_chain,
            "net_chain": net_chain,
            "message_hash": message_hash,
            "message_bytes": message_bytes,
            "metadata": metadata
        }
        return message

    @classmethod
    def _evaluate_broadcast(cls, long_buffer, message):
        if (not long_buffer.has(message['self_chain']) and
                not long_buffer.has(message['net_chain'])):
            # check if this is the first message on the net or an unsynced peer
            if message['self_chain'] != message['net_chain']:
                message['error_code'] = OUT_OF_ORDER_BOTH
                return message
            else:
                message['error_code'] = OUT_OF_ORDER_SELF
                return message
        elif not long_buffer.has(message['net_chain']):
            message['error_code'] = OUT_OF_ORDER_NET
            return message
        elif not (long_buffer.has(message['self_chain'])):
            message['error_code'] = OUT_OF_ORDER_SELF
            return message
        return message

    @classmethod
    def _evaluate_direct(cls, long_buffer, message):
        # no need to check net_chain for a direct message
        if not (long_buffer.has(message['self_chain'])):
            message['error_code'] = OUT_OF_ORDER_SELF
            return message
        return message

    @classmethod
    def _unpack_message(cls, message_bytes):
        int_ts, self_chain, net_chain, speaker, body = struct.unpack(MESSAGE_PACKET_FORMAT, message_bytes)
        speaker, body = cls._unpad_message_speaker_and_body(speaker, body)
        return int_ts, self_chain, net_chain, speaker, body

    @classmethod
    def _unpack_getdata_message(cls, message_bytes):
        int_ts, self_chain, net_chain, speaker, body, padding = struct.unpack(GETDATA_MESSAGE_PACKET_FORMAT, message_bytes)
        return int_ts, self_chain, net_chain, speaker, body

    @classmethod
    def _unpad_message_speaker_and_body(cls, speaker, body):
        # remove padding from speaker
        for index, byte in enumerate(speaker):
            if byte == '\x00':
                speaker = speaker[0:index]
                break

        # remove padding from body
        for index, byte in enumerate(body):
            if byte == '\x00':
                body = body[0:index]
                break

        try:
            return speaker.decode('ascii'), body.decode('utf-8')

        except Exception:
            return speaker, body

    @classmethod
    def _pad(cls, text, size):
        return text.ljust(size, "\x00")

    @classmethod
    def _check_for_ascii_encoding(cls, handle):
        for c in handle:
            if ord(c) not in range(0,127):
                return False
        return True

    @classmethod
    def _ts_range(cls):
        current_ts = int(time.time())
        return range(current_ts - TS_ACCEPTABLE_SKEW, current_ts + TS_ACCEPTABLE_SKEW)

    @classmethod
    def gen_hash(cls, message_bytes):
        return hashlib.sha256(message_bytes).digest()

    def get_message_bytes(self, peer=None):
        command = self.command
        speaker = Message._pad(self.speaker, MAX_SPEAKER_SIZE)

        # let's generate the self_chain value from the last message or set it to zero if 
        # this is the first message
        if command == DIRECT:
            self_chain = self.state.get_handle_self_chain(peer.handles[0])
            net_chain = EMPTY_CHAIN
        elif command == BROADCAST:
            self_chain = self.state.get_broadcast_self_chain()
            net_chain = self.state.get_net_chain()
        elif command == IGNORE:
            self_chain = net_chain = EMPTY_CHAIN
        elif command == GETDATA:
            self_chain = net_chain = EMPTY_CHAIN

        self.self_chain = self_chain
        self.net_chain = net_chain

        if command in [IGNORE, GETDATA]:
            body = self.body
        else:
            body = self.body.encode('utf-8')

        message_bytes = struct.pack(MESSAGE_PACKET_FORMAT,
            self.timestamp,
            self_chain,
            net_chain,
            speaker.encode('ascii'),
            body)
        return message_bytes

    def compute_message_hash(self):
        if self.message_hash is None:
            if self.message_bytes is not None:
                self.message_hash = Message.gen_hash(self.message_bytes)
                return self.message_hash
            else:
                return None
        else:
            return self.message_hash

    def log_outgoing(self, peer):
        logging.info(OUTGOING_MESSAGE_LOGGING_FORMAT % (
            peer.address,
            peer.port,
            peer.handles[0],
            COMMAND_LABELS[self.command],
            self.body,
            self.bounces,
            binascii.hexlify(self.compute_message_hash())))

    def log_rubbish(self, peer):
        logging.info(OUTGOING_MESSAGE_LOGGING_FORMAT % (peer.address,
                                                        peer.port,
                                                        peer.handles[0],
                                                        COMMAND_LABELS[self.command],
                                                        "<rubbish>",
                                                        self.bounces,
                                                        binascii.hexlify(self.message_hash)))

    def log_incoming(self, peer):
        try:
            logging.info(INCOMING_MESSAGE_LOGGING_FORMAT % (peer.address,
                peer.port,
                peer.handles[0],
                COMMAND_LABELS[self.command],
                self.body,
                self.bounces,
                binascii.hexlify(self.message_hash)))
        except Exception, ex:
            logging.info("unable to log incoming message")

    def log_incoming_getdata(self, peer):
        try:
            logging.info(INCOMING_MESSAGE_LOGGING_FORMAT % (peer.address,
                                                            peer.port,
                                                            peer.handles[0],
                                                            COMMAND_LABELS[self.command],
                                                            binascii.hexlify(self.body),
                                                            self.bounces,
                                                            binascii.hexlify(self.message_hash)))
        except Exception, ex:
            logging.info("unable to log incoming message")
    def retry(self, requesting_peer):
        logging.debug("Can't retry a message that isn't DIRECT or BROADCAST")

        return

    @classmethod
    def in_time_window(cls, timestamp):
        return timestamp in cls._ts_range()
